#include "SkillImplementationWrapper.h"

namespace armarx
{
    namespace skills::detail
    {
        void SkillImplementationWrapper::reset()
        {
            ARMARX_CHECK_NOT_NULL(skill);
            std::lock_guard l(skillStatusMutex);
            statusUpdate.status = skills::provider::dto::Execution::Status::Idle;
            statusUpdate.data = nullptr;
            skill->reset();
        }

        void SkillImplementationWrapper::execute()
        {
            ARMARX_CHECK_NOT_NULL(skill);
            std::lock_guard l(executingMutex);

            const std::string skillName = skill->description.skillName;
            ARMARX_INFO_S << "Executing skill: " << skillName;

            // get params and setup variables
            auto& aron_params = statusUpdate.usedParameterization;

            try
            {
                // Check params
                if (not(aron_params) && skill->description.acceptedType)
                {
                    throw armarx::LocalException("The Skill '" + skillName + "' requires a type but no params are NULL.");
                }

                // set scheduled
                {
                    std::lock_guard l(skillStatusMutex);
                    statusUpdate.status = skills::provider::dto::Execution::Status::Scheduled;
                    updateStatus();
                }


                // execute
                {
                    std::lock_guard l(skillStatusMutex);
                    statusUpdate.status = skills::provider::dto::Execution::Status::Running;
                    updateStatus();
                }

                auto execution_callback = [&](const aron::data::DictPtr& update)
                {
                    statusUpdate.data = update;
                    updateStatus();
                };
                auto result = skill->execute(aron_params, execution_callback);

                switch (result)
                {
                    case skills::Skill::Status::Failed:
                        throw armarx::LocalException("The Skill '" + skillName + "' failed during execution.");
                        break;
                    case skills::Skill::Status::TimeoutReached:
                        throw armarx::LocalException("The Skill '" + skillName + "' reached timeout during execution.");
                        break;
                    case skills::Skill::Status::Stopped:
                    {
                        std::lock_guard l(skillStatusMutex);
                        statusUpdate.status = skills::provider::dto::Execution::Status::Aborted;
                        updateStatus();
                        break;
                    }
                    default:
                    {
                        std::lock_guard l(skillStatusMutex);
                        statusUpdate.status = skills::provider::dto::Execution::Status::Succeeded;
                        updateStatus();
                    }
                }
            }
            catch (const std::exception& ex)
            {
                ARMARX_WARNING_S << "Skill " << skillName << " died with exception:\n" << ex.what();

                std::lock_guard l(skillStatusMutex);
                statusUpdate.status = skills::provider::dto::Execution::Status::Failed;
                updateStatus();
            }
        }

        void SkillImplementationWrapper::updateStatus() const
        {
            auto& callbackInterface = statusUpdate.usedCallbackInterface;

            if (callbackInterface)
            {
                callbackInterface->updateStatusForSkill(statusUpdate.toIce());
            }
        }
    }
}
